import functools

from ..utils.query_util import parse_sql_where, parse_sql_update, execute


def update(sql: str):
    def deco(fun):
        @functools.wraps(fun)
        def wrapper(data, **kwargs):
            fun(data)
            in_sql, query_dict = parse_sql_update(sql, data)
            in_sql = parse_sql_where(in_sql, kwargs)
            execute(in_sql, {**query_dict, **kwargs})

        return wrapper

    return deco



import inspect

from pypadb.table.base_table import BaseTable
from pypadb.utils.inspect_util import camel2snake


class TableConfigurer:
    __slots__ = ['tables']

    tables: dict[str, BaseTable]

    def __init__(self):
        self.tables = {}

    def __getattr__(self, item: str) -> BaseTable:
        return self.tables[item]

    def __getitem__(self, item: str) -> BaseTable:
        return self.tables[item]

    def init_tables(self, module_=None, escape_upper: bool = False, **kwargs):
        if inspect.ismodule(module_):
            module_dict: dict = {}
            module_member = inspect.getmembers(module_)
            magic_all = [i[1] for i in module_member if i[0] == '__all__'][0]
            if magic_all:
                module_member = [i for i in module_member if i[0] in magic_all]
            for i in module_member:
                member_name = camel2snake(i[0])
                module_dict[member_name] = i[1]
            self.tables = {**self.tables, **{key: BaseTable(key, module_dict[key]) for key in module_dict}}
        if kwargs:
            self.tables = {**self.tables, **{key: BaseTable(key, kwargs[key]) for key in kwargs}}


tables: TableConfigurer = TableConfigurer()



from typing import Any

from ..connection_pool import cursor_type, connection
from ..utils.conditions import Limit
from ..utils.enums import QueryModeEnum


def query(sql: str, kwargs: dict, data_type: Any, model: QueryModeEnum):
    result_set, last_row_id = execute(sql, kwargs, cursor_type())
    if result_set is None:
        return None

    if model == QueryModeEnum.One:
        return data_type(**result_set[0]), last_row_id
    elif model == QueryModeEnum.Many:
        return [data_type(**i) for i in result_set], last_row_id


def execute(sql: str, kwargs: dict = None, cursor_type=None) -> tuple[list, int]:
    conn = connection()
    with conn:
        cur = conn.cursor(cursor_type) if cursor_type else conn.cursor()
        cur.execute(sql, kwargs)
        conn.commit()
        return cur.fetchall(), cur.lastrowid


def parse_sql_batch(columns: list, data_list: list) -> tuple[str, dict]:
    query_dict: dict = {}
    query_sql: str = ''
    for i, v in enumerate(data_list):
        data_dict: dict = v.dict()
        query_dict = {**query_dict, **{f'{arg}{i}': data_dict.get(arg) for arg in columns}}
        query_sql += '('
        for item in columns:
            query_sql += f'%({item}{i})s,'
        query_sql = query_sql.rstrip(',') + '),'
    return query_sql.rstrip(','), query_dict


def parse_sql_where(sql: str, conditions: dict, limit: Limit = None) -> str:
    if conditions:
        sql += ' where'
        for key in conditions:
            sql += f' {key}=%({key})s and'
        sql = sql.rstrip(' and')
    return sql + str(limit) if limit else sql


def parse_sql_update(sql, entity) -> tuple[str, dict]:
    data_dict = entity.dict()
    query_dict: dict = {}
    salt = '_a_salt_that_never_duplicate'
    sql += ' set'
    for key in data_dict:
        if data_dict[key] or data_dict[key] == 0:
            sql += f' {key}=%({key}{salt})s,'
            query_dict[f'{key}{salt}'] = data_dict[key]
    return sql.rstrip(','), query_dict



import abc
from typing import Any

from ..utils.enums import LikeEnum, QueryModeEnum


class Limit:
    __slots__ = ['start', 'count']

    start: int
    count: int

    def __init__(self, start: int, count: int = None):
        self.start = start
        self.count = count

    def __str__(self):
        if self.count is None:
            return f' limit {self.start}'
        return f' limit {self.start}, {self.count}'


class Like:
    __slots__ = ['sql', 'value', 'column']

    sql: str
    column: str
    value: str

    def __init__(self, column: str, value, like_type: LikeEnum):
        self.column = column
        self.value = value
        self.sql = f' {column} like'
        if like_type == LikeEnum.L_Like:
            self.sql += f' concat(\'%%\',%({column})s)'
        elif like_type == LikeEnum.R_Like:
            self.sql += f' concat(%({column})s,\'%%\')'
        elif like_type == LikeEnum.ALL_Like:
            self.sql += f' concat(\'%%\',%({column})s,\'%%\')'

    def __str__(self):
        return self.sql


def extra(column: list, data_property: str, method):
    """
    :param column: table A[column[0]] mapping table B[column[1]]
    :param data_property: table A property name
    :param method: query method
    """

    def e(data_entity):
        if isinstance(data_entity, list):
            for entity in data_entity:
                entity.__setattr__(data_property,
                                   method(**{column[1]: data_entity.__getattribute__(column[0])}))
            return data_entity
        data_entity.__setattr__(data_property,
                                method(**{column[1]: data_entity.__getattribute__(column[0])}))
        return data_entity

    return e



import inspect
from typing import Any, Optional

from pydantic import BaseModel

from ..exception import NotCallable


class Argument(BaseModel):
    name: str
    type: Optional[type]
    default: Any


def arg_list(fun) -> list[Argument]:
    if not callable(fun):
        raise NotCallable(fun)
    return [Argument(name=tp[0],
                     type=tp[1].annotation if tp[1].annotation != inspect._empty else None,
                     default=tp[1].default if tp[1].default != inspect._empty else None)
            for tp in inspect.signature(fun).parameters.items()]


def returns_type(fun) -> type:
    a = inspect.signature(fun).return_annotation
    return a if a != inspect._empty else None


def camel2snake(source: str):
    res = source[0].lower()
    for i in source[1:]:
        if i.isupper():
            res += '_'
        res += i.lower()
    return res



from typing import Union

from dbutils.pooled_db import PooledDB, PooledDedicatedDBConnection, PooledSharedDBConnection

from ..exception import PoolUninitializedException

_POOL: PooledDB = None
_Cursor = None


def _init_pool(cursor, **data: dict):
    global _POOL, _Cursor
    _POOL = PooledDB(**data)
    _Cursor = cursor


def connection() -> Union[PooledSharedDBConnection, PooledDedicatedDBConnection]:
    if not _POOL:
        raise PoolUninitializedException('connection pool uninitialized')
    return _POOL.connection()


def cursor_type():
    return _Cursor
